/*************************************************************************

   This software was written by Fred Cerou, Inria, France 
   (cerou@sophia.inria.fr) in collaboration with Pr. Rene Carmona, 
   Princeton University, USA (rcarmona@chelsea.princeton.edu). Most 
   of the code was written at Princeton University, while Fred Cerou
   was visiting the Civil Engineering and Operations Research Dpt. 

   Read file flow1.doc for more information.

   Copyright (C) 1997 Rene Carmona, Fred Cerou.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.  This software is provided "as is" without express or
   implied warranty.

**************************************************************************/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>     /* Get standard string definitions. */
#include <X11/xpm.h> /* xpm after Intrinsic to avoid redefinition of Pixel */

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/AsciiText.h>

#define TRUE  1
#define FALSE 0

/*
 *   Begining of user defined parameters
 */


/* size of the display window */
#define XMAX 200
#define YMAX 200

/* size of the displayed state space area */
#define FXMAX 800.0
#define FXMIN -800.0
#define FYMAX 800.0
#define FYMIN -800.0

/* radius of the circle where to put the drifters at time=0 
   in the case we don't want a square (see SQUARE below) */
#define RADIUS 100.0

/* define SHEAR for shear flow case */
/*#define SHEAR*/

/* constant x-speed for shear flow */
#define WSHEAR 10.0

/* define SQUARE to start with dirfters in a square 
   for given size (on a circle otherwise) */
#define SQUARE 150

/* define GNUPLOT to save several states in gnuplot format */
/*#define GNUPLOT*/

/* number of states saved in gnuplot format */
#define NBSAVE 3

/* define DRAW_FIELD to show modulus and direction of the 
   velocity field on the whole region of space */
/*#define DRAW_FIELD*/

/* used to rescale  the colors of the field modulus. the best 
   value to choose depends on the parameters in flow1.data */
#define COLORSCALE 8


/* define ADD_POINTS to add drifters for keeping a "continous"
   curve. not use if SQUARE is defined */
/*#define ADD_POINTS */

/* define SAVE_AUX to save some auxiliary quantities in 
   flow1.out.aux */
#define SAVE_AUX

/* number of simulation steps between each display on the 
   screen */
#define IDRAW 3

/* define SAVE_IMAGES to store the images in pixmap files */
/*#define SAVE_IMAGES*/

/* define quantile to compute a given level quantile for 
   the distance to the origin. results stored in flow1.out.q */
/*#define QUANTILE 0.95*/

/* number of simulation steps between each quantile computation */
#define IQUANT 1

/* define HISTO to compute histograms of the distance to the origin,
   results stored in flow1.out.h */
/*#define HISTO*/

/* number of simulation steps between each histogram computation */
#define IHISTO 20

/* number of intervals for the histogram */
#define NHISTO 30

/* max value for the histogram */
#define HISTMAX 700.0

/* colors used to display the drifters */
#define COLOR1 255
#define COLOR2 70
#define COLOR3 130
#define COLOR4 195

/* random seed  */
#define SEED 2344309

/*
 * End of user defined parameters
 */

const double DELTAX = (FXMAX - FXMIN)/XMAX;
const double DELTAY = (FYMAX - FYMIN)/YMAX;

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif


double ran1(long *idum);
double fastnorm(long *idum, double mean, double sd, double (*func)());

struct DrifterStruct
{
  double x,y;
  short color;
  struct DrifterStruct *next,*prev;
};

typedef struct DrifterStruct DrifterType;

double sqr(double x)
{
  return(x*x);
}

int drcompare(DrifterType **dr1, DrifterType **dr2)
{
  double r1,r2;

  r1 = (*dr1)->x * (*dr1)->x + (*dr1)->y * (*dr1)->y;
  r2 = (*dr2)->x * (*dr2)->x + (*dr2)->y * (*dr2)->y;

  if(r1>r2) return 1;
  if(r1<r2) return -1;
  return 0;
}

void main(int argc, char **argv)
{

  /*
   * Variables related to stream function
   */

  int NbFourierCirc, NbFourierCircModes,NbFourierModes;
  double RminFourier, RmaxFourier, EpsFourier, zFourier;
  double *k1Fourier, *k2Fourier;
  double *aFourier, *bFourier;
  double *alphaFourier, *sigmaFourier;

  /*
   * Number and current positions of drifters
   */

  int NbDrifters;

  DrifterType *dr1,*dr2,*RootDrifter,**drtab=NULL;

  /* 
   * Field at drifter's position
   */

  double vxDrifter, vyDrifter;

  /*
   * General settings
   */

  int nbTime, iTime;

  double finalTime, currentTime, stepTime;


  /*
   * Loop idices
   */

  int i,j,k,jj,kk;

  /*
   * Auxiliary variables
   */

  double KX, asinKX, bcosKX, radius, radius1, radius2, angle, coeff, x, y;

  double delta, deltamax;

  double Diam, Diamx, Diamy, xmin, xmax, ymax, ymin;

  double d1,d2,dmax;

  int index,*histotab;

  /*
   * Files
   */

  FILE *inFile, **outFile, *auxFile, *qfile, *hfile, *hauxfile;
  char outFileName[30];
  int outTime;

  /*
   * XWindows
   */

  Widget toplevel, box, dessin, dessin1, dessin2;
  Pixmap Xpix=(Pixmap)NULL;
  Pixmap Xpix1=(Pixmap)NULL;
  Colormap cmap;
  GC noir,blanc,copie,cols[256];
  FILE *ColorFile;
  XColor couleur[256];
  int red,green,blue;
  unsigned long valuemask; 
  XSetWindowAttributes winattr;
  Arg arg[2];
  int mycolor;
  XpmAttributes attributes;
  Display   *ecran;
  XtAppContext app_con;

  /*
   * Random seed
   */

  long idum;

  /*
   * Init X
   */

  toplevel = XtAppInitialize(&app_con,"top",NULL,ZERO,&argc,argv,
			     NULL , NULL, ZERO);

  
  ecran=XtDisplay (toplevel);

  XtSetArg(arg[0], XtNheight, YMAX);
  XtSetArg(arg[1], XtNwidth, XMAX);

  box = XtCreateManagedWidget("box",boxWidgetClass,toplevel,NULL, ZERO);

  dessin  = XtCreateManagedWidget("dessin",simpleWidgetClass,box,arg,TWO);

#ifdef DRAW_FIELD
  dessin1 = XtCreateManagedWidget("dessin",simpleWidgetClass,box,arg,TWO);
  dessin2 = XtCreateManagedWidget("dessin",simpleWidgetClass,box,arg,TWO);
#endif

  XtRealizeWidget(toplevel);

  cmap = XCreateColormap(ecran,RootWindow(ecran,DefaultScreen(ecran)),
			 DefaultVisual(ecran,DefaultScreen(ecran)),AllocAll);
  ColorFile = fopen("palette","r");
  for(i=0;i<256;i++)
    {
      fscanf(ColorFile,"%d %d %d",&red,&green,&blue);
      couleur[i].flags = DoRed|DoGreen|DoBlue;
      couleur[i].red = 65535*red/255;
      couleur[i].green = 65535*green/255;
      couleur[i].blue = 65535*blue/255;
      couleur[i].pixel = i;
    }
  couleur[1].red = 0;
  couleur[1].green = 0;
  couleur[1].blue = 0;
  couleur[0].red = 65535;
  couleur[0].green = 65535;
  couleur[0].blue = 65535;
  couleur[255].red = 50000;
  couleur[255].green = 50000;
  couleur[255].blue = 65535;
  close(ColorFile);
  XStoreColors(ecran,cmap,couleur,256);
  valuemask = CWColormap;
  winattr.colormap = cmap;
  XChangeWindowAttributes(ecran,XtWindow(toplevel),valuemask,&winattr);
  XChangeWindowAttributes(ecran,XtWindow(dessin),valuemask,&winattr);

#ifdef DRAW_FIELD
  XChangeWindowAttributes(ecran,XtWindow(dessin1),valuemask,&winattr);
  XChangeWindowAttributes(ecran,XtWindow(dessin2),valuemask,&winattr);
#endif

  for(i=0;i<256;i++)
    {
      cols[i] = XCreateGC (ecran,XtWindow(dessin),
			   0, NULL);
      XSetFunction (ecran,cols[i],GXcopy);
      XSetForeground(ecran,cols[i],(unsigned long)i);

    }


  noir = cols[1];
  blanc = cols[0];

  copie = XCreateGC (ecran,XtWindow(dessin),0, NULL);
  XSetFunction (ecran,copie,GXcopy);

  attributes.valuemask = XpmColormap;
  attributes.colormap = cmap;

  Xpix = XCreatePixmap(ecran,XtWindow(dessin),XMAX,YMAX,
		       DefaultDepthOfScreen(XtScreen(dessin)));

#ifdef DRAW_FIELD
  Xpix1 = XCreatePixmap(ecran,XtWindow(dessin),XMAX,YMAX,
		       DefaultDepthOfScreen(XtScreen(dessin)));
#endif

  XFillRectangle(ecran,Xpix,noir,0,0,XMAX,YMAX);

  /*
   * Initialize the random numbers generator
   */
  
  idum = - SEED;
  if(idum==0) idum = -63472;
  if(idum>0) idum = -idum;
  ran1(&idum);


  /* 
   * Read data in the input file
   */

  inFile = fopen("flow1.data","r");

  fscanf(inFile,"%d",&NbFourierCirc);

  fscanf(inFile,"%d",&NbFourierCircModes);

  fscanf(inFile,"%lf%lf",&RminFourier,&RmaxFourier);

  fscanf(inFile,"%lf%lf",&EpsFourier,&zFourier);

  if (NbFourierCirc>0)
    NbFourierModes = NbFourierCircModes*NbFourierCirc;
  else
    NbFourierModes = NbFourierCircModes;


  /*
   * Initialization of the k's and the O-U processes
   */

  k1Fourier    = (double *)malloc(NbFourierModes*sizeof(double));
  k2Fourier    = (double *)malloc(NbFourierModes*sizeof(double));
  aFourier     = (double *)malloc(NbFourierModes*sizeof(double));
  bFourier     = (double *)malloc(NbFourierModes*sizeof(double));
  alphaFourier = (double *)malloc(NbFourierModes*sizeof(double));
  sigmaFourier = (double *)malloc(NbFourierModes*sizeof(double));
  outFile      = (FILE **)malloc(4*sizeof(FILE *));

#ifndef SHEAR
  if(NbFourierCirc==0) /* Random modes */
    {
      for(j=0;j<NbFourierCircModes;j++)
	{
	  angle = 2.0*M_PI*ran1(&idum); 
	  radius = RminFourier + (RmaxFourier - RminFourier)*ran1(&idum);

	  k1Fourier[j] = radius*cos(angle);
	  k2Fourier[j] = radius*sin(angle);
	  
	  sigmaFourier[j] = pow(radius,-2.0-EpsFourier);
	  alphaFourier[j] = pow(radius,zFourier);
	}
    }
  else
    {     
      for(i=0;i<NbFourierCirc;i++)
	{
	  radius = RmaxFourier -
	    i*(RmaxFourier - RminFourier)/(NbFourierCirc-1);
	  radius1 = RmaxFourier -
	    (i-0.5)*(RmaxFourier - RminFourier)/(NbFourierCirc-1);
	  radius2 = RmaxFourier -
	    (i+0.5)*(RmaxFourier - RminFourier)/(NbFourierCirc-1);

	  for(j=0;j<NbFourierCircModes;j++)
	    {
	      angle = 2*M_PI*j/NbFourierCircModes + 2*M_PI/(i+1);
	      index = i*NbFourierCircModes+j;
	      k1Fourier[index] = radius*cos(angle);
	      k2Fourier[index] = radius*sin(angle);
	      
	      sigmaFourier[index] = pow(radius,-2.0-EpsFourier);
	      alphaFourier[index] = pow(radius,zFourier);

	      /* Normalize by the elementary domain in Fourier space */
	      sigmaFourier[index] *= 
		sqrt((radius1*radius1 - radius2*radius2)
		     *(2*M_PI/NbFourierCircModes));
	    }
	}
    }
#else /*Shear flow case */
  if((NbFourierModes%2)!=0) NbFourierModes--;
  for(i=0;i<NbFourierModes/2;i++)
    {
      radius = -RmaxFourier + 
	2*i*(RmaxFourier - RminFourier)/NbFourierModes;
      k1Fourier[i] = radius;
      k2Fourier[i] = 0.0;
      sigmaFourier[i] = pow(-radius,-2.0-EpsFourier);
      alphaFourier[i] = pow(-radius,zFourier);

      /* Normalize by the elementary domain in Fourier space */
      sigmaFourier[i] *= 2*(RmaxFourier - RminFourier)/NbFourierModes;
    }
  for(i=0;i<NbFourierModes/2;i++)
    {
      radius = RmaxFourier -
	2*i*(RmaxFourier - RminFourier)/NbFourierModes;
      k1Fourier[NbFourierModes/2+i] = radius;
      k2Fourier[NbFourierModes/2+i] = 0.0;
      sigmaFourier[NbFourierModes/2+i] = pow(radius,-2.0-EpsFourier);
      alphaFourier[NbFourierModes/2+i] = pow(radius,zFourier);

      /* Normalize by the elementary domain in Fourier space */
      sigmaFourier[i] *= sqrt(2*(RmaxFourier - RminFourier)/NbFourierModes);
    }
#endif

  fscanf(inFile,"%d%lf",&nbTime,&stepTime);

  fscanf(inFile,"%d",&NbDrifters);

  close(inFile);


  auxFile = fopen("flow1.fourier","w");
  fprintf(auxFile,"#Nb Fourier Modes : %d\n",NbFourierModes);
  for(index=0;index<NbFourierModes;index++)
    {
      fprintf(auxFile,"%lf %lf %lf %lf\n",k1Fourier[index],k2Fourier[index],
	      sigmaFourier[index],alphaFourier[index]);
    }
  fflush(auxFile);
  close(auxFile);

  /*
   * Initial values of O-U proceses according to the invariant law
   */

  for(i=0;i<NbFourierModes;i++)
    {
      aFourier[i] = fastnorm(&idum,0.0E0,
			     sigmaFourier[i]/sqrt(2*alphaFourier[i]),ran1);
      bFourier[i] = fastnorm(&idum,0.0E0,
			     sigmaFourier[i]/sqrt(2*alphaFourier[i]),ran1);
    }

  /*
   * modify the sigma's and alpha's according to time step
   */

  for(i=0;i<NbFourierModes;i++) 
    {
      sigmaFourier[i] = sigmaFourier[i]*
	sqrt((1.0 - exp(-2.0*stepTime*alphaFourier[i]))/(2.0*alphaFourier[i]));

      alphaFourier[i] = exp(-alphaFourier[i]*stepTime);
    }

  /*
   * Initial position of the drifters
   */

  RootDrifter = (DrifterType *)malloc(sizeof(DrifterType));
  dr1 = RootDrifter;

  for(i=1;i<NbDrifters;i++)
    {
      dr2 = (DrifterType *)malloc(sizeof(DrifterType));
      dr2->prev = dr1;
      dr1->next = dr2;
      dr1 = dr2;
    }

  RootDrifter->prev = dr2;
  dr2->next = RootDrifter;

  dr1 = RootDrifter;

#ifdef SQUARE

  for(i=0;i<sqrt(NbDrifters);i++)
  for(j=0;j<sqrt(NbDrifters);j++)
    {
      dr1->x = -SQUARE/2.0 + SQUARE*i/sqrt(NbDrifters);
      dr1->y = -SQUARE/2.0 + SQUARE*j/sqrt(NbDrifters);
      if((dr1->x>0.0)&&(dr1->y>=0.0)) dr1->color = COLOR1;
      if((dr1->x<=0.0)&&(dr1->y>0.0)) dr1->color = COLOR2;
      if((dr1->x<0.0)&&(dr1->y<=0.0)) dr1->color = COLOR3;
      if((dr1->x>=0.0)&&(dr1->y<0.0)) dr1->color = COLOR4;
      dr1 = dr1->next;
    }
#else
  for(i=0;i<NbDrifters;i++)
    {
      dr1->x = RADIUS*cos(2*i*M_PI/NbDrifters);
      dr1->y = RADIUS*sin(2*i*M_PI/NbDrifters);
      if(4*i/NbDrifters==0) dr1->color = COLOR1;
      if(4*i/NbDrifters==1) dr1->color = COLOR2;
      if(4*i/NbDrifters==2) dr1->color = COLOR3;
      if(4*i/NbDrifters==3) dr1->color = COLOR4;
      dr1 = dr1->next;
    }
#endif

  deltamax = RADIUS*2.0*sin(M_PI/NbDrifters);

  outTime = nbTime/NBSAVE;

#ifdef SAVE_AUX
  auxFile = fopen("flow1.out.aux","w");
  fprintf(auxFile,"# deltamax: %lf\n",deltamax);
#endif

#ifdef QUANTILE
  qfile = fopen("flow1.out.q","w");
#endif

#ifdef HISTO
  hfile = fopen("flow1.out.h","w");
  hauxfile=fopen("flow1.out.h.aux","w");
  histotab = (int *)malloc(NbDrifters*sizeof(int));
#endif

  /* 
   * Main loop in time
   */

  currentTime = 0.0;

  for(iTime=0;iTime<nbTime;iTime++)
    {
      /*
       * Write output, if needed
       */

      if (iTime%outTime==0)
	{
	  printf("%d\n",iTime/outTime);

#ifdef GNUPLOT
	  for(i=0;i<4;i++)
		{
		  sprintf(outFileName,"flow1.out.%d.%d",iTime/outTime,i+1);

		  outFile[i] = fopen(outFileName,"w");

		  fprintf(outFile[i],"# time = %lf\n", currentTime);
		  fprintf(outFile[i],"# drifters = %d\n",NbDrifters);
		  fprintf(outFile[i],"# color = %d\n",i);
		}

	  dr1 = RootDrifter;
	  for(i=0;i<NbDrifters;i++)
	    {
	      if(dr1->color==COLOR1) kk = 1;
	      if(dr1->color==COLOR2) kk = 2;
	      if(dr1->color==COLOR3) kk = 3;
	      if(dr1->color==COLOR4) kk = 4;
	      fprintf(outFile[kk],"%lf\t%lf\n",dr1->x,dr1->y);
	      dr1 = dr1->next;
	    }
	  fprintf(outFile[3],"%lf\t%lf\n",
		  RootDrifter->x,RootDrifter->y);

	  for(i=0;i<4;i++)
	    close(outFile[i]);
#endif
	}

      currentTime += stepTime;

#ifdef QUANTILE
      if((iTime%IQUANT)==0)
	{
	  if(drtab) free(drtab);
	  drtab = (DrifterType **)malloc(NbDrifters*sizeof(DrifterType *));
	  dr1 = RootDrifter;
	  drtab[0] = dr1;
	  for(i=1;i<NbDrifters;i++)
	    {
	      dr1 = dr1->next;
	      drtab[i] = dr1;
	    } 
	  qsort(drtab,NbDrifters,sizeof(DrifterType *),drcompare);
	  dr1 = drtab[NbDrifters-1];
	  dr2 = drtab[(int)(NbDrifters*QUANTILE)];
	  d1 = sqrt(dr1->x * dr1->x + dr1->y * dr1->y);
	  d2 = sqrt(dr2->x * dr2->x + dr2->y * dr2->y);
	  fprintf(qfile,"%lf %lf %lf\n",currentTime,d1,d2);
	}
#endif

#ifdef HISTO
      if ((iTime%IHISTO)==0)
	{
	  dr1 = RootDrifter;
	  dmax = sqrt(dr1->x * dr1->x + dr1->y * dr1->y);
	  for(i=1;i<NbDrifters;i++)
	    {
	      d1 = sqrt(dr1->x * dr1->x + dr1->y * dr1->y);
	      if(d1>dmax) dmax=d1;
	      dr1 = dr1->next;
	    }
	  for(i=0;i<NHISTO;i++) histotab[i] = 0;
	  dr1 = RootDrifter;
	  for(i=0;i<NbDrifters;i++)
	    {
	      d1 = sqrt(dr1->x * dr1->x + dr1->y * dr1->y);
	      histotab[(int)(d1*(NHISTO-1)/HISTMAX)]++;
	      dr1 = dr1->next;
	    }
	  for(i=0;i<NHISTO;i++)
	    fprintf(hfile,"%lf\n",((double) histotab[i])/NbDrifters);
	  fprintf(hfile,"\n");

	  fprintf(hauxfile,"%d %lf\n",NbDrifters,dmax);
	}
#endif

#ifdef SAVE_AUX
	  
      /*
       * Compute the diameter for \|.\|_{\infty} 
       */

      dr1 = RootDrifter;

      xmax = dr1->x;
      ymax = dr1->y;
      xmin = dr1->x;
      ymin = dr1->y;

      dr1 = dr1->next;

      for(i=1;i<NbDrifters;i++)
	{
	  if(dr1->x > xmax) xmax = dr1->x;
	  if(dr1->x < xmin) xmin = dr1->x;
	  if(dr1->y > ymax) ymax = dr1->y;
	  if(dr1->y < ymin) ymin = dr1->y;

	  dr1 = dr1->next;
	}

      Diamx = xmax - xmin;
      Diamy = ymax - ymin;
      Diam = (Diamx>Diamy) ? Diamx : Diamy;

      dr1 = RootDrifter;
      kk = 0;
      dmax = sqrt(dr1->x * dr1->x + dr1->y * dr1->y);
      for(i=1;i<NbDrifters;i++)
	{
	  dr1 = dr1->next;
	  d1 = sqrt(dr1->x * dr1->x + dr1->y * dr1->y);
	  if(d1>dmax) {dmax=d1;kk=i;}
	}

      /*
       * Write some auxiliary values...
       */

      fprintf(auxFile,"%lf\t%d\t%lf\t%lf\t%lf\t%d\n",
	      currentTime,NbDrifters,Diam,
	      RootDrifter->x,RootDrifter->y,kk);
      fflush(auxFile);
#endif

      /*
       * Update the O-U processes
       */

      for(i=0;i<NbFourierModes;i++)
	{
	  aFourier[i] = fastnorm(&idum,aFourier[i]*alphaFourier[i],
				 sigmaFourier[i],ran1);
	      
	  bFourier[i] = fastnorm(&idum,bFourier[i]*alphaFourier[i],
				 sigmaFourier[i],ran1);
	}

      if((iTime%IDRAW)==0)
	{
	  XFillRectangle(ecran,Xpix,noir,0,0,XMAX,YMAX);
	  printf("%d\n",iTime);
	  
	  /*
	   * Draw Field
	   */
	  
#ifdef DRAW_FIELD
	  for(jj=0;jj<XMAX;jj++)
	    for(kk=0;kk<YMAX;kk++)
	      {
		x= FXMIN + jj*DELTAX;
		y= FYMAX - kk*DELTAY;
		
		/*
		 * Update v(t,x)
		 */
		
		vxDrifter = 0.0;
		vyDrifter = 0.0;
		
		for(i=0;i<NbFourierModes;i++)
		  {
		    KX = k1Fourier[i] * x + k2Fourier[i] * y;

		    asinKX = aFourier[i]*sin(KX);
		    bcosKX = bFourier[i]*cos(KX);

#ifndef SHEAR
		    vxDrifter += k2Fourier[i]*(bcosKX - asinKX);
#endif
		    vyDrifter += k1Fourier[i]*(asinKX - bcosKX);
		  }

#ifdef SHEAR
		vxDrifter = WSHEAR;
#endif
		/* Intensity */
		mycolor = 2 +
		  sqrt(vxDrifter*vxDrifter+vyDrifter*vyDrifter)/COLORSCALE;
		if (mycolor>190) mycolor=190;
		XDrawPoint(ecran,Xpix,cols[mycolor],jj,kk);

		/* Direction*/
		mycolor = 2+(M_PI+atan2(vyDrifter,vxDrifter))*250/(2*M_PI);
		XDrawPoint(ecran,Xpix1,cols[mycolor],jj,kk);
		
	      }
	  XCopyArea(ecran,Xpix,XtWindow(dessin1),copie,0,0,XMAX,YMAX,0,0);
#ifdef SAVE_IMAGES
	  sprintf(outFileName,"flow1mod.%d.xpm",iTime);
	  XpmWriteFileFromPixmap(ecran,outFileName,Xpix,NULL,&attributes);
#endif
	  XCopyArea(ecran,Xpix1,XtWindow(dessin2),copie,0,0,XMAX,YMAX,0,0);
#ifdef SAVE_IMAGES
	  sprintf(outFileName,"flow1dir.%d.xpm",iTime);
	  XpmWriteFileFromPixmap(ecran,outFileName,Xpix1,NULL,&attributes);
#endif
	  XFlush(ecran);
#endif      
	}

      XFillRectangle(ecran,Xpix,noir,0,0,XMAX,YMAX);
      
      dr1 = RootDrifter;
      for(j=0;j<NbDrifters;j++)
	{

	  /*
	   * Update v(t,x)
	   */

	  vxDrifter = 0.0;
	  vyDrifter = 0.0;

	  for(i=0;i<NbFourierModes;i++)
	    {
	      KX = k1Fourier[i] * dr1->x + k2Fourier[i] * dr1->y;
	  
	      asinKX = aFourier[i]*sin(KX);
	      bcosKX = bFourier[i]*cos(KX);

#ifndef SHEAR
	      vxDrifter += k2Fourier[i]*(bcosKX - asinKX);
#endif
	      vyDrifter += k1Fourier[i]*(asinKX - bcosKX);
	    }

#ifdef SHEAR
	  vxDrifter = WSHEAR;
#endif

	  /* 
	   * Update drifter's position
	   */

	  dr1->x += vxDrifter*stepTime;
	  dr1->y += vyDrifter*stepTime;

	  dr1 = dr1->next;
	}

      if((iTime%IDRAW)==0)
	{
	  /*
	   * Draw Point
	   */
	  dr1 = RootDrifter;
	  for(j=0;j<NbDrifters;j++)
	    {
	      XDrawPoint(ecran,Xpix,cols[dr1->color],
			 (int) (((dr1->x)-(FXMIN))/DELTAX),
			 (int) ((FYMAX-(dr1->y))/DELTAY));
	      dr1 = dr1->next;
	    }
	  XCopyArea(ecran,Xpix,XtWindow(dessin),copie,0,0,XMAX,YMAX,0,0);
	  XFlush(ecran);
	  
#ifdef SAVE_IMAGES
	  sprintf(outFileName,"flow1.%d.xpm",iTime);
	  XpmWriteFileFromPixmap(ecran,outFileName,Xpix,NULL,&attributes);
#endif
	}

      /*
       * Add Drifters
       */
      
      dr1 = RootDrifter;

#ifndef SQUARE
#ifdef ADD_POINTS
      for(j=0;j<=NbDrifters;j++)
	{
	  delta = sqrt(sqr(dr1->x - dr1->next->x) 
                         + sqr(dr1->y - dr1->next->y));

	  if(delta>deltamax)
	    {
	      dr2 = (DrifterType *)malloc(sizeof(DrifterType));
	      dr2->color = dr1->color;
	      dr2->x = 0.5*(dr1->x + dr1->next->x);
	      dr2->y = 0.5*(dr1->y + dr1->next->y);
	      dr2->prev = dr1;
	      dr2->next = dr1->next;
	      dr2->next->prev = dr2;
	      dr1->next = dr2;
	      NbDrifters++;
	    }
	  
	  dr1 = dr1->next;
	}
#endif
#endif
      
    }

#ifdef GNUPLOT
  for(i=0;i<4;i++)
    {
      sprintf(outFileName,"flow1.out.%d.%d",iTime/outTime,i+1);
      
      outFile[i] = fopen(outFileName,"w");
      
      fprintf(outFile[i],"# time = %lf\n", currentTime);
      fprintf(outFile[i],"# drifters = %d\n",NbDrifters);
    }

  dr1 = RootDrifter;
  for(i=0;i<NbDrifters;i++)
    {
      if(dr1->color==COLOR1) kk = 1;
      if(dr1->color==COLOR2) kk = 2;
      if(dr1->color==COLOR3) kk = 3;
      if(dr1->color==COLOR4) kk = 4;
      fprintf(outFile[kk],"%lf\t%lf\n",dr1->x,dr1->y);
      dr1 = dr1->next;
    }
  fprintf(outFile[3],"%lf\t%lf\n",
	  RootDrifter->x,RootDrifter->y);
  
  for(i=0;i<4;i++)
    close(outFile[i]);
#endif

#ifdef QUANTILE
  close(qfile);
#endif 

#ifdef HISTO
  close(hfile);
  close(hauxfile);
#endif

#ifdef SAVE_AUX    
  close(auxFile);
#endif  

  dr1 = RootDrifter;
  for(i=0;i<NbDrifters;i++)
    {
      dr2 = dr1->next;
      free(dr1);
      dr1 = dr2;
    }
  free(k1Fourier);
  free(k2Fourier);
  free(aFourier);
  free(bFourier);
  free(alphaFourier);
  free(sigmaFourier);
  free(outFile);

}
  
